
/*------------------------------------------------------------------
|																	|
|	Line															|
|																	|
 ------------------------------------------------------------------*/

class Line
{
	towns              = null;
	stations           = null;
	mail_stations      = null;	
	depot              = null;
	vehicles           = null;
	date_last_vehicle  = 0;
	failed             = false;
	passenger_cargo_id = 0;
	mail_cargo_id      = 0;
	pending_vehicles   = 0;
	new_location       = null;
	mail_location       = null;
	success            = 0;
	retry_build        = false;
	num_retries	       = 0;	
	town_exit          = null;
	line_len           = 0;
	state              = 0;
	pf 				   = 0;		
	connected		   = false;
	constructor(cargo_id, m_cargo_id, town_pair) {
		if (town_pair.len() == 2){
			towns     = town_pair;
		} else {
			AILog.Info("2 towns needed to create a Line");
			return null;
		}
		
		stations           = [[], []];
		mail_stations      = [null, null];
		depot              = null;
		vehicles           = [];
		failed             = false;
		date_last_vehicle  = 0;
		passenger_cargo_id = cargo_id;
		mail_cargo_id      = m_cargo_id;
		new_location       = [0,0];
		mail_location      = [0,0];
		success            = [false, false];
		pending_vehicles   = 0;
		retry_build        = false;
		num_retries        = 0;
		town_exit          = [0, 0];
		line_len           = 0;
		state              = 0;
		connected		   = false;
		pf 			       = PathFinder();
	}
	function AddVehicles();
	function EstimateBusesNeeded(station0, station1);
	function BuildDepot();
	function CreateNewLine(town_pair);
	function ManageVehicles();
	function Controller();
	function BuildBusStop(tile);
}

function Line::Controller()
{
	
	switch (state){
		case 0:
			AILog.Info("State " + state);
			if (PlanStations()){
				state = 1;
			} else {
				state = 10;
			}	
			break;
		case 1:
			if (Cfo.HasMoney(15000)){
				AILog.Info("State " + state);
				BuildStations();
				if((success[0] == RoadBuilder.SUCCESS) && (success[1] == RoadBuilder.SUCCESS)
					&& town_exit[0] && town_exit[1] ) {
					state = 2;
					num_retries = 0;
				} else if ((success[0] == RoadBuilder.REPLAN) || (success[1] == RoadBuilder.REPLAN)){
					state = 0;
					num_retries++;
					if (num_retries > 3){
						state = 10;
						AILog.Info("Max retries: go to state " + state);
					}
				} else if ((success[0] == RoadBuilder.RETRY_LATER) || (success[1] == RoadBuilder.RETRY_LATER)){
					state = 1;
				}	
				else state = 10;
			}
			break;
		case 2:
			if (Cfo.HasMoney(3000)) {
				AILog.Info("State " + state);
				if (BuildDepot()) {
					AILog.Info("Depot built " + depot);
					state = 3;
					num_retries = 0;
				}
				else {
					AILog.Info("Failed to build depot");
				}
			}
			break;
		case 3:
			if (Cfo.HasMoney(15000)) {
				AILog.Info("State " + state);
				if (PlanRoute()){
					state = 4;
				}	
			}	
			break;
		case 4:
			if (Cfo.HasMoney(16000)) {
				AILog.Info("State " + state);
				if (BuildRoute()){
					state = 5;
				} else {
					if (pf.retry_find) { 
						if (num_retries < 3) {
							num_retries++;
							state = 3;
						} else {
							state = 10;
						}
					}			
				}		
			}	
			break;
		case 5:
			if (Cfo.HasMoney(15000)){
				AILog.Info("State " + state);
				if (AddVehicles()){
					state = 6;
				}	
			}
			break;
		case 6:
			if (Cfo.HasMoney(15000)){
//				AILog.Info("State " + state);
				if (ManageVehicles()){
					state = 7;
					num_retries = 0;
				}
			}
			break;
		case 7:
			if (Cfo.HasMoney(40000)){
				AILog.Info("State " + state);
				if (PlanMailTruckStops()){
					state = 8;
				} else {
					state = 10;
				}	
			}	
			break;
		case 8:
			if (Cfo.HasMoney(10000)){
				AILog.Info("State " + state);
				BuildMailTruckStops();
				if((success[0]) && (success[1])){
					state = 9;
				} else if (retry_build){
					state = 7;
					num_retries++;
					if (num_retries > 3){
						state = 10;
						AILog.Info("Max retries: go to state " + state);
					}
				}
			}
			break;	
		case 9:
			if (Cfo.HasMoney(5000)){
				AILog.Info("State " + state);
				if (AddMailTrucks()){
					state = 10;
				}	
			}
			break;	
		default:
			break;
	}
	// Pay back unused money
	Cfo.ManageLoan();
	return (state < 6);
}

function Line::PlanStations()
{
	local town_idx     = 0;
	AILog.Info("Plan stations, line from: " + AITown.GetName(towns[0]) + " to " + AITown.GetName(towns[1]) + " " +  AICompany.GetBankBalance(AICompany.COMPANY_SELF));
	if(towns && (towns.len() == 2))
	{
		foreach (town in towns){
			if (!success[town_idx]){
				new_location[town_idx] = TownManager.FindLineBusStopLocation(town, passenger_cargo_id, false);
			}	
			town_idx++;
		}
	}else {
		AILog.Info("Create new line failed: towns array not filled");
	}	
	if (new_location[0] && new_location[1]){
		AILog.Info("Plan stations: locations found");
		return true;
	} else {
		AILog.Info("Plan stations: locations not found");
		return false;
	}	
}

function Line::PlanMailTruckStops()
{
	local town_idx     = 0;
	AILog.Info("Plan mail truck stop: " + AITown.GetName(towns[0]) + " to " + AITown.GetName(towns[1]) + " " +  AICompany.GetBankBalance(AICompany.COMPANY_SELF));
	if(towns && (towns.len() == 2))
	{
		foreach (town in towns){
			if (!mail_location[town_idx]){
				mail_location[town_idx] = TownManager.FindMailTruckStopLocation(stations[town_idx][0], mail_cargo_id, false);
			}	
			town_idx++;
		}
	}else {
		AILog.Info("Plan mail truck stop failed: towns array not filled");
	}	
	if (mail_location[0] && mail_location[1]){
		success[0] = false;
		success[1] = false;
		return true;
	} else {
		return false;
	}	
}


/*------------------------------------------------------------------*/
function Line::BuildBusStop(tile)
{
	local found = false;
	local adjacentTiles = Tile.GetAdjacentTiles(tile);
	local rb      = RoadBuilder();
	local success = RoadBuilder.REPLAN;
	
	AILog.Info("BuildBusStop: " + AITown.GetName(towns[0]) + " to " + AITown.GetName(towns[1]) + " " +  AICompany.GetBankBalance(AICompany.COMPANY_SELF));
	for(local tile2 = adjacentTiles.Begin(); 
			adjacentTiles.HasNext() && !found; 
			tile2 = adjacentTiles.Next()) {
		if(AIRoad.IsRoadTile(tile2) && !AITile.GetSlope(tile2)) {
			success = rb.BuildRoadStation(tile, tile2, true);
			found = (success == RoadBuilder.SUCCESS);
		} 
	}
	
	return success;	
}


/*------------------------------------------------------------------*/
function Line::BuildStations()
{
	local connected = false;
	local town_idx     = 0;
	
	foreach (town in towns){
		if (new_location[town_idx] && success[town_idx] != RoadBuilder.SUCCESS){
			AILog.Info("BuildStation new loc " + town_idx + "  " + new_location[town_idx]);
			success[town_idx] = BuildBusStop(new_location[town_idx]);
			if (success[town_idx] == RoadBuilder.SUCCESS) {
				AILog.Info("busstop " + town_idx + " built");
				stations[town_idx].append (new_location[town_idx]);	
				town_exit[town_idx] = AIRoad.GetRoadStationFrontTile(new_location[town_idx]);			
				AILog.Info("BuildStation town exit " + town_idx + "  " + town_exit[town_idx]);
				if (!town_exit[town_idx]) {
					AILog.Info("Find exit " + town_idx + " failed");
					return;
				}			
			} else{
				AILog.Info("New line: failed building busstop " + AIError.GetLastErrorString());
				return;
			}
		}
		town_idx++;
	}
}	

/*------------------------------------------------------------------*/
function Line::BuildMailTruckStop(tile)
{
	local found = false;
	local success = false;

	local adjacentTiles = Tile.GetAdjacentTiles(tile);

	AILog.Info("BuildMailTruckStop: " + AITown.GetName(towns[0]) + " to " + AITown.GetName(towns[1]) + " " +  AICompany.GetBankBalance(AICompany.COMPANY_SELF));
	for(local tile2 = adjacentTiles.Begin(); 
			adjacentTiles.HasNext() && !found; 
			tile2 = adjacentTiles.Next()) {
		if(AIRoad.IsRoadTile(tile2) && !AITile.GetSlope(tile2)) {
			local rb      = RoadBuilder();
			success = rb.BuildRoadStation(tile, tile2, false);
			found = (success == RoadBuilder.SUCCESS);
			retry_build = (success == RoadBuilder.REPLAN);
			success = found;
		} else {
			retry_build = true;
		}
	}

	return success;	
}

/*------------------------------------------------------------------*/
function Line::BuildMailTruckStops()
{
	local connected = false;

	local town_idx     = 0;

	foreach (town in towns){
		if (mail_location[town_idx] && !success[town_idx]){
			if (!mail_stations[town_idx]){
				success[town_idx] = BuildMailTruckStop(mail_location[town_idx]);
				if (success[town_idx]) {
					AILog.Info("mail truck stop " + town_idx + " built");
					mail_stations[town_idx]= mail_location[town_idx];
				} else{
					AILog.Info("New line: failed building mail truckstop " + AIError.GetLastErrorString());
					break;
				}
			}
		}
		town_idx++;
	}
	return (mail_stations[0] && mail_stations[1]);

}	

/*------------------------------------------------------------------*/
function Line::PlanRoute()
{
	local connected = true;

	connected = pf.Plan(town_exit[0], town_exit[1]); 
	if (!connected) {
		AILog.Info("Plan Route failed "+ AIError.GetLastErrorString());
	}

	return connected;
}

/*------------------------------------------------------------------*/
function Line::BuildRoute()
{
	local conn;
	local rb = RoadBuilder();
	

	if (!connected){
		connected = (rb.BuildPath(pf.path) == RoadBuilder.SUCCESS);
	}	
	conn = connected;
	if (connected){
		line_len = rb.path_len;
		// make sure the depot and station are connected
		conn = pf.Plan(AIRoad.GetRoadStationFrontTile(stations[0][0]), AIRoad.GetRoadDepotFrontTile(depot)) ;
		if (conn) {
			conn = (rb.BuildPath(pf.path) == RoadBuilder.SUCCESS);
		}	
		// make sure the depot and town_exit are connected for a nicer route
		if(!conn){
			AILog.Info("Build or plan route to depot failed"+ AIError.GetLastErrorString());
		}
	} 
	else {
		AILog.Info("Build route failed"+ AIError.GetLastErrorString());
	}

	return conn;
}

/*------------------------------------------------------------------*/
function Line::AddVehicles()
{
	local bus_started = false;
	if(this.stations[0].len() != this.stations[1].len()) {
		AILog.Info("AddVehicles nof stations incorrect");
		return null;
	}
	local bus_model = null;
	local min_reliability = 85;
	local new_bus;
	local i;
	AILog.Info("Add vehicles to line from: " + AITown.GetName(towns[0]) + " to " + AITown.GetName(towns[1]) + " " +  AICompany.GetBankBalance(AICompany.COMPANY_SELF));
	local engine_list = AIEngineList(AIVehicle.VT_ROAD);
	
	engine_list.Valuate(AIEngine.GetRoadType);
	engine_list.KeepValue(AIRoad.ROADTYPE_ROAD);
	
	local balance = AICompany.GetBankBalance(AICompany.COMPANY_SELF);
	engine_list.Valuate(AIEngine.GetPrice);
	engine_list.KeepBelowValue(balance);

	engine_list.Valuate(AIEngine.GetCargoType)
	engine_list.KeepValue(passenger_cargo_id); 

	engine_list.Valuate(AIEngine.GetCapacity)
	engine_list.KeepTop(1);
	
	pending_vehicles = EstimateBusesNeeded(this.stations[0][this.stations[0].len()-1],
											this.stations[1][this.stations[1].len()-1]);

	if (!engine_list.Count()) {
		AILog.Info("BuyBuses: failed to find a bus to build");
		return false;
	}

	bus_model = engine_list.Begin();

	// Buy buses

	for(local bus_idx  = 0; bus_idx < 2; ++bus_idx)
	{
		local town_idx = 0;	
		new_bus = AIVehicle.BuildVehicle(this.depot, bus_model);
//		local er = AIError.GetLastError();
		if (AIVehicle.IsValidVehicle(new_bus)) { 
			this.vehicles.append(new_bus);
			local bus = this.vehicles[bus_idx]; 

			AIOrder.AppendOrder(bus, this.depot, AIOrder.AIOF_SERVICE_IF_NEEDED);
			AIOrder.AppendOrder(bus, this.stations[town_idx][0], AIOrder.AIOF_NONE); 
			town_idx = (town_idx + 1) % 2;
			AIOrder.AppendOrder(bus, this.stations[town_idx][0], AIOrder.AIOF_NONE);
			AIVehicle.StartStopVehicle(new_bus);
			bus_started  = true;
		}
		else {
			AILog.Info("Buy vehicles failed "+ AIError.GetLastErrorString() + " " + this.depot);
		}
		
	}
	AILog.Info("Buses pending:" + pending_vehicles);
	return (bus_started);

}


/*------------------------------------------------------------------*/
function Line::AddMailTrucks()
{
	local truck_started = false;
	if(!this.mail_stations[0] || !this.mail_stations[1]) {
		AILog.Info("AddMailTrucks nof stations incorrect");
		return null;
	}
	local truck_model = null;
	local min_reliability = 70;
	local new_truck;
	local i;
	AILog.Info("Add mail trucks to line from: " + AITown.GetName(towns[0]) + " to " + AITown.GetName(towns[1]) + " " +  AICompany.GetBankBalance(AICompany.COMPANY_SELF));
	local engine_list = AIEngineList(AIVehicle.VT_ROAD);
	
	engine_list.Valuate(AIEngine.GetRoadType);
	engine_list.KeepValue(AIRoad.ROADTYPE_ROAD);
	
	local balance = AICompany.GetBankBalance(AICompany.COMPANY_SELF);
	engine_list.Valuate(AIEngine.GetPrice);
	engine_list.KeepBelowValue(balance);

	engine_list.Valuate(AIEngine.GetCargoType)
	engine_list.KeepValue(mail_cargo_id); 

	engine_list.Valuate(AIEngine.GetCapacity)
	engine_list.KeepTop(1);
	
	pending_vehicles = 4;  // todo : tune this value

	if (!engine_list.Count()) {
		AILog.Info("Buy trucks: failed to find a mail truck to build");
		return false;
	}

	truck_model = engine_list.Begin();

	// Buy buses

	for(local truck_idx  = 0; truck_idx < 4; ++truck_idx)
	{
		local town_idx = 0;	
		new_truck = AIVehicle.BuildVehicle(this.depot, truck_model);
//		local er = AIError.GetLastError();
		if (AIVehicle.IsValidVehicle(new_truck)) { 
			local truck = new_truck; 

			AIOrder.AppendOrder(truck, this.depot, AIOrder.AIOF_SERVICE_IF_NEEDED);
			AIOrder.AppendOrder(truck, this.mail_stations[town_idx], AIOrder.AIOF_NONE); 
			town_idx = (town_idx + 1) % 2;
			AIOrder.AppendOrder(truck, this.mail_stations[town_idx], AIOrder.AIOF_NONE);
			AIVehicle.StartStopVehicle(new_truck);
			truck_started  = true;
		}
		else {
			AILog.Info("Buy vehicles failed "+ AIError.GetLastErrorString() + " " + this.depot);
		}
		
	}
	AILog.Info("Trucks pending:" + pending_vehicles);

	return (truck_started);
}


/*------------------------------------------------------------------*/
function Line::EstimateBusesNeeded(station0, station1)
{

	if(!station0 || !station1)
		return 0;
	
	local acceptance = AITile.GetCargoAcceptance(station0, 
												passenger_cargo_id, 1, 1, 
												AIStation.GetCoverageRadius (AIStation.STATION_BUS_STOP)) + 
												AITile.GetCargoAcceptance(station1, 
												passenger_cargo_id, 1, 1, 
												AIStation.GetCoverageRadius (AIStation.STATION_BUS_STOP)) ;
	local distance = AIMap.DistanceManhattan( station0, station1 );

	AILog.Info("EstimateBusesNeeded: distance:    " + distance + 
	 										 " acceptance:  " + acceptance);

	local num_bus = 2 + (acceptance / 35) * (distance / 35);
	
	if (num_bus > 25) num_bus = 25;
	AILog.Info("Buy " + num_bus + " buses");
	return num_bus;
}



/*------------------------------------------------------------------*/
function Line::BuildDepot()
{
	local tl     = AITileList();
	local tile   = stations[0][0];
	local x      = AIMap.GetTileX(tile);
	local y      = AIMap.GetTileY(tile);
	local aitile = 0;
	local found  = false;
	
	tl.AddRectangle(AIMap.GetTileIndex(x-10,y-10), AIMap.GetTileIndex(x+10,y+10));
	if (tl.Count())
	{
		if (tl.Count()) {
			/* find all tiles that are next to a road tile */

			tl.Valuate(AIRoad.GetNeighbourRoadCount);	
			tl.KeepAboveValue(0);
		}
		if (tl.Count()) {
			/* find all tiles that are not road */
			tl.Valuate(AIRoad.IsRoadTile);	
			tl.KeepValue(0);
		}
		if (tl.Count()) {
			/* find all tiles that are not sloped */
			tl.Valuate(AITile.GetSlope);	
			tl.KeepValue(0);
		}
		if (tl.Count()) {
			tl.Valuate(AITile.GetDistanceManhattanToTile, AITown.GetLocation(towns[1]));
			tl.Sort(AITileList.SORT_BY_VALUE, true);
		
			for (aitile = tl.Begin();tl.HasNext() && !found; 
				aitile = tl.Next()) {
				local adjacentTiles = Tile.GetAdjacentTiles(aitile);

				/*Loop through all adjacent tiles*/

				for(local tile2 = adjacentTiles.Begin(); 
					adjacentTiles.HasNext() && !found; 
					tile2 = adjacentTiles.Next()) {
					if(AIRoad.IsRoadTile(tile2) && !AITile.GetSlope(tile2) ) {
						if (!AIRoad.IsRoadTile(tile2+tile2-aitile)) {
							if(!AIRoad.IsRoadStationTile(aitile)) {
								found = AITile.IsBuildable(aitile);
								if (!found)	{
									found = AITile.DemolishTile(aitile);
								}	
								if (found) {
									local rb  = RoadBuilder();
									local success = false;
									success = rb.BuildDepot(aitile, tile2);
									found = (success == RoadBuilder.SUCCESS); 
								}
							}
						}
					}
				}
				if (found) {
					break;
				}
			}
		}
	}
	if (found) {
		depot = aitile;
		return aitile;	
	}
	else {
		depot = null;
		return null;
	}
}



/*------------------------------------------------------------------*/
function Line::ManageVehicles()
{
	local max_veh_on_line;
	local max_veh_reached = false;
	
	foreach (station in this.stations[0]){
		local curr_station = AIStation.GetStationID(station);
		local veh_list_station = AIVehicleList_Station(curr_station);
		local veh_on_line = veh_list_station.Count();		

		max_veh_on_line  = 5 + this.line_len / 7;
		if (veh_on_line < max_veh_on_line) {
			local second_station = AIStationList_Vehicle(veh_list_station.Begin()).Begin();
			local waiting = AIStation.GetCargoWaiting(curr_station,
												  	  passenger_cargo_id) +
														AIStation.GetCargoWaiting(AIStation.GetStationID(second_station),
														passenger_cargo_id);
			local rating0 = AIStation.GetCargoRating(curr_station, passenger_cargo_id);
			local rating1 = AIStation.GetCargoRating(second_station, passenger_cargo_id);

			if ((waiting > 45) || (rating0 < 75) || (rating1 < 75)) {
				if (AIDate.GetCurrentDate() - this.date_last_vehicle > 50) {
					if (veh_list_station.Begin() && this.depot) {
						this.pending_vehicles = this.pending_vehicles + waiting/45;
					}
				}
			}
		}
//		AILog.Info("Manage vehicles: pending , max " + this.pending_vehicles + ", " + max_veh_on_line + ", " + this.vehicles.len());
		if (this.vehicles.len() + this.pending_vehicles >= max_veh_on_line){
			this.pending_vehicles = max_veh_on_line - this.vehicles.len();
			
		}	
			
		if ((AIDate.GetCurrentDate() - this.date_last_vehicle) > 10){
			while (this.pending_vehicles) {
				if (this.vehicles.len()){
					local new_veh = AIVehicle.CloneVehicle (this.depot, 
															this.vehicles[0], true);			
					if (AIVehicle.IsValidVehicle(new_veh)) { 										
						this.vehicles.append(new_veh);				
						this.date_last_vehicle = AIDate.GetCurrentDate();
						AIVehicle.StartStopVehicle(new_veh);
						if (this.vehicles.len() >= max_veh_on_line){
							max_veh_reached = true;
							AILog.Info("Manage vehicles: max veh reached " + max_veh_on_line);
						}	
					}
				} else {
					AILog.Info(this.name + "vehicles[0] is false");
				}			
					
				this.pending_vehicles = this.pending_vehicles - 1;
			}
		}
	}
	return (max_veh_reached);
}



/*------------------------------------------------------------------
function ConvoyAI::ManageVehicles()
{
	local list = AIVehicleList();
	list.Valuate(AIVehicle.GetAge);
	list.KeepAboveValue(700);
	list.Valuate(AIVehicle.GetProfitLastYear);

	for (local i = list.Begin(); list.HasNext(); i = list.Next()) {
		local profit = list.GetValue(i);

		if (profit < -100) {
			local station_lst = AIStationList_Vehicle(i);
			local station0 = null;
			local station1 = null;
			if (station_lst) {
				station0 = station_lst.Begin(); 
				
				if (station_lst.HasNext()){
					station1 = station_lst.Next();
				}
			}	
			if (station0 && station1) {
				local rating_st0 = AIStation.GetCargoRating(station0, passenger_cargo_id);
				local rating_st1 = AIStation.GetCargoRating(station1, passenger_cargo_id);			
				if ((rating_st0 >= 40) && (rating_st1 >= 40)){
					/* Send the vehicle to depot if we didn't do so yet 
					if (!this.vehicle_to_depot.rawin(i) || this.vehicle_to_depot.rawget(i) != true) {
						AILog.Info(this.name + ": [INFO] Sending " + i + " to depot as	profit is: " + profit );
						AIVehicle.SendVehicleToDepot(i);
						this.vehicle_to_depot.rawset(i, true);
					}
				}
			}	
		}
		/* Try to sell it over and over till it really is in the depot 
		if (this.vehicle_to_depot.rawin(i) && this.vehicle_to_depot.rawget(i) == true) {
			if (AIVehicle.SellVehicle(i)) {
				AILog.Info(this.name + ": [INFO] Selling " + i + " as it finally is in a depot.");

				this.vehicle_to_depot.rawdelete(i);
			}
		}
	}
}
*/
