
class RailBuilder
{
	SUCCESS        = 0;
	RETRY_LATER    = 1;
	REPLAN         = 2;
	FAIL           = 3;
	path_len       = 0;
	rail_list	   = [];	
	constructor() {
		SUCCESS        = 0;
		RETRY_LATER    = 1;
		REPLAN         = 2;
		FAIL           = 3;
		path_len       = 0;
	}
	function BuildPath( path );
}


function RailBuilder::BuildPath( path )
{
	local build_success = false;
	local retry_find    = false;
	path_len     	    = 0;	
	local prev = null;
	local prevprev = null;
	
	while (path) {
		path_len = path_len + 1;	
		rail_list.append(path.GetTile());
		if (prevprev != null) {
			local retry_build = false;
			local build_retries = 0;	 
			do {
				retry_build = false;
				if (AIMap.DistanceManhattan(prev, path.GetTile()) > 1) {					
					if (AITunnel.GetOtherTunnelEnd(prev) == path.GetTile()) {
						build_success = AITunnel.BuildTunnel(AIVehicle.VT_RAIL, prev);
						AILog.Info("Build Tunnel");
					} else {
						local bridge_list = AIBridgeList_Length(AIMap.DistanceManhattan(path.GetTile(), prev) + 1);
						bridge_list.Valuate(AIBridge.GetMaxSpeed);
						bridge_list.Sort(AIAbstractList.SORT_BY_VALUE, false);
						build_success = AIBridge.BuildBridge(AIVehicle.VT_RAIL, bridge_list.Begin(), prev, path.GetTile());
						AILog.Info("Build Bridge");
					}
					if (build_success){
						prevprev = prev;
						prev = path.GetTile();
						path = path.GetParent();
					}
				}	
				else {
					build_success = AIRail.BuildRail(prevprev, prev, path.GetTile());
//					AILog.Info("Build Rail");
				}
				if (!build_success) {
					local err = AIError.GetLastError();
					switch (err) {
						/* ignore these errors */
						case AIError.ERR_NONE: 	
						case AIError.ERR_ALREADY_BUILT: 
//							AILog.Info("Build Rail:"+ AIError.GetLastErrorString());
							build_success = true;
						break;
						case AIError.ERR_NOT_ENOUGH_CASH:  
							AILog.Info("Build Rail failed (no cash)"+ AIError.GetLastErrorString());
							return RETRY_LATER;
						break;
						case AIError.ERR_VEHICLE_IN_THE_WAY: 
							AILog.Info("Build Rail failed (retry)"+ AIError.GetLastErrorString());	
							AIController.Sleep(1);
						break;
						case AIError.ERR_AREA_NOT_CLEAR: 	
							AILog.Info("Build Rail failed (demolish + retry)"+ AIError.GetLastErrorString());	
							if ((!AIRail.IsRailStationTile (path.GetTile())) &&
								(!AIRail.IsRailDepotTile (path.GetTile())) &&
								(!AIRoad.IsRoadStationTile (path.GetTile())) &&
								(!AIRoad.IsRoadDepotTile (path.GetTile()))
								){
								retry_find = !AITile.DemolishTile(path.GetTile());
							} else {
								retry_find = true;
							}	
							break;
						default:
							AILog.Info("Build Rail failed (replan)"+ AIError.GetLastErrorString());	
							retry_find = true;
							break;
					}
				} 	
				build_retries++;					
			} while (!build_success && (build_retries < 80) && !retry_find);
  		}
		if (path) {
			prevprev = prev;
			prev = path.GetTile();
			path = path.GetParent();
		}
	}
	
	if (build_success)
		return SUCCESS;
	else if (retry_find) 
		return REPLAN;
	else
		return FAIL;	
}



function RailBuilder::BuildRailStation( station )
{
	local adjacent = AIStation.STATION_NEW;
	local success = false;


	
	success = AIRail.BuildRailStation(station._location, AIRail.RAILTRACK_NE_SW, 2, 5, adjacent);
	if (!success){
		local err = AIError.GetLastError();
		switch (err) {
			/* ignore these errors */
			case AIError.ERR_NONE: 	
			case AIError.ERR_ALREADY_BUILT: 
				AILog.Info("Build Rail:" + AIError.GetLastErrorString());
				return SUCCESS;
			break;
			case AIError.ERR_NOT_ENOUGH_CASH:  
				AILog.Info("Build Rail station failed (no cash)" + AIError.GetLastErrorString());
				return RETRY_LATER;
			break;
			default:
				AILog.Info("Build Rail station failed " + AIError.GetLastErrorString());
				return FAIL;
			break;
		}
	} else {
		station.SetSize(2,5);
		return SUCCESS;
	}	
}

function RailBuilder::BuildDepot(from, front, to, dep)
{
	local rb  = RailBuilder();
	local success;
	if (from && front && dep){
		success = AIRail.BuildRail(from, front, dep);
	}	
	if (to && front && dep)
		AIRail.BuildRail(dep, front, to);
	if (!success){
		local err = AIError.GetLastError();
		switch (err) {
			/* ignore these errors */
			case AIError.ERR_NONE: 	
			case AIError.ERR_ALREADY_BUILT: 
				AILog.Info("Build Depot Rail:" + AIError.GetLastErrorString());
			break;
			case AIError.ERR_NOT_ENOUGH_CASH:  
				AILog.Info("Build depot Rail failed (no cash)" + AIError.GetLastErrorString());
				return RETRY_LATER;
			break;
			default:
				AILog.Info("Build Depot Rail failed " + AIError.GetLastErrorString());
				return FAIL;
			break;
		}
	}	

	success = AIRail.BuildRailDepot(dep, front)
	
	if (!success){
		local err = AIError.GetLastError();
		switch (err) {
			/* ignore these errors */
			case AIError.ERR_NONE: 	
			case AIError.ERR_ALREADY_BUILT: 
				AILog.Info("Build Depot:" + AIError.GetLastErrorString());
				return SUCCESS;
			break;
			case AIError.ERR_NOT_ENOUGH_CASH:  
				AILog.Info("Build Depot failed (no cash)" + AIError.GetLastErrorString());
				return RETRY_LATER;
			break;
			default:
				AILog.Info("Build Rail station failed " + AIError.GetLastErrorString());
				return FAIL;
			break;
		}
	} else {
		return SUCCESS;
	}	
}

function RailBuilder::IsRailCrossing(tile)
{
	local tracks = AIRail.GetRailTracks(tile);
	local nof_bits = 0;
	local mask = 1;
	
	for (local bit_idx = 0; bit_idx < 6; bit_idx++){
		if (tracks & mask) nof_bits++;
		if (nof_bits == 2){
			if (((tracks & (AIRail.RAILTRACK_NE_SE | AIRail.RAILTRACK_NW_SW)) != (AIRail.RAILTRACK_NE_SE | AIRail.RAILTRACK_NW_SW)) &&
				((tracks & (AIRail.RAILTRACK_NW_NE | AIRail.RAILTRACK_SW_SE)) != (AIRail.RAILTRACK_NW_NE | AIRail.RAILTRACK_SW_SE)))
			return true;
		}	
		mask = mask << 1;
	}
	return false;
}


function RailBuilder::MakeSwitchpoint(path1, path2)
{
	local tile1x, tile2x, tile1y, tile2y, tile1prv, tile2prv, tile1, tile2;
	local tile1pprev = null, tile2pprev = null;
	local mapsize_x = AIMap.GetMapSizeX() - 1;	
	local raillist1 = [];
	local raillist2 = [];
	local idx = 0; 
	
	local tmp_path = path1;
	while (tmp_path) {
		raillist1.append(tmp_path.GetTile());
		tmp_path = tmp_path.GetParent();
	}
	tmp_path = path2;
	while (tmp_path) {
		raillist2.append(tmp_path.GetTile());
		tmp_path = tmp_path.GetParent();
	}
	local len1 = raillist1.len();
	local len2 = raillist2.len();	
	AIRail.BuildSignal(raillist1[1], raillist1[0], AIRail.SIGNALTYPE_PBS);
	AIRail.BuildSignal(raillist2[1], raillist2[0], AIRail.SIGNALTYPE_PBS);
	AIRail.BuildSignal(raillist1[len1 - 2], raillist1[len1 - 1], AIRail.SIGNALTYPE_PBS);
	AIRail.BuildSignal(raillist2[len2 - 2], raillist2[len2 - 1], AIRail.SIGNALTYPE_PBS);
	
	local pf = RailPathFinder(true);	
	if ((raillist1.len() > 15) && (raillist2.len() > 15)){
		local tmp1_idx = 0;
		local tmp2_idx = 0;
		local success = false;
		local source_tiles = [];
		local dest_tiles = [];
		while ((tmp1_idx < 15) && !success){
			source_tiles.append([raillist1[tmp1_idx + 1], raillist1[tmp1_idx]]);
			tmp1_idx++;
		}	
		while ((tmp2_idx < 15) && !success){
			dest_tiles.append([raillist2[tmp2_idx],raillist2[tmp2_idx + 1]]);
			tmp2_idx++;
		}	
		pf.Plan(source_tiles, dest_tiles, [], 750);
		success = (BuildPath(pf.path) == SUCCESS);
	}	
	if ((raillist1.len() > 15) && (raillist2.len() > 15)){
		local tmp1_idx = 0;
		local tmp2_idx = 0;
		local success = false;
		local source_tiles = [];
		local dest_tiles = [];
		while ((tmp1_idx < 15) && !success){
			dest_tiles.append([raillist1[tmp1_idx], raillist1[tmp1_idx + 1]]);
			tmp1_idx++;
		}	
		while ((tmp2_idx < 15) && !success){
			source_tiles.append([raillist2[tmp2_idx + 1],raillist2[tmp2_idx]]);
			tmp2_idx++;
		}	
		pf.Plan(source_tiles, dest_tiles, [], 750);
		success = (BuildPath(pf.path) == SUCCESS);
	}
	if ((raillist1.len() > 15) && (raillist2.len() > 15)){
		local tmp1_idx = 0;
		local tmp2_idx = 0;
		local success = false;
		local source_tiles = [];
		local dest_tiles = [];
		while ((tmp1_idx < 15) && !success){
			source_tiles.append([raillist1[len1 - tmp1_idx - 2], raillist1[len1 - tmp1_idx - 1]]);
			tmp1_idx++;
		}	
		while ((tmp2_idx < 15) && !success){
			dest_tiles.append([raillist2[len2 - tmp2_idx - 1],raillist2[len2 - tmp2_idx - 2]]);
			tmp2_idx++;
		}	
		pf.Plan(source_tiles, dest_tiles, [], 750);
		success = (BuildPath(pf.path) == SUCCESS);
	}
	if ((raillist1.len() > 15) && (raillist2.len() > 15)){
		local tmp1_idx = 0;
		local tmp2_idx = 0;
		local success = false;
		local source_tiles = [];
		local dest_tiles = [];
		while ((tmp1_idx < 15) && !success){
			dest_tiles.append([raillist1[len1 - tmp1_idx - 1], raillist1[len1 - tmp1_idx - 2]]);
			tmp1_idx++;
		}	
		while ((tmp2_idx < 15) && !success){
			source_tiles.append([raillist2[len2 - tmp2_idx - 2],raillist2[len2 - tmp2_idx - 1]]);
			tmp2_idx++;
		}	
		pf.Plan(source_tiles, dest_tiles, [], 750);
		success = (BuildPath(pf.path) == SUCCESS);
	}
	tile1prv = path1.GetTile();
	tile2prv = path2.GetTile();
	path1  = path1.GetParent();
	path2  = path2.GetParent();

	while (path1 != null) {
		idx++;
		tile1 = path1.GetTile();
		local tracks1 = AIRail.GetRailTracks(tile1);

//		if ((idx>15) && (tracks1 & (AIRail.RAILTRACK_NE_SW | AIRail.RAILTRACK_NW_SE))) {
		if (idx>15) { 
			local match = false;
			tile1x = tile1 & mapsize_x;
			tile1y = tile1 & ~mapsize_x;
		
			local path2_sav = path2;
			local tile2pp_sav = tile2pprev;
			local tile2prv_sav = tile2prv;
			// search second path until x or y matches
			if (path2) tile2 = path2.GetTile();

			while (((tile2y = (tile2 & ~mapsize_x)) != tile1y) 
				 && ((tile2x = (tile2 & mapsize_x)) != tile1x)
				 &&(path2 != null)){
					tile2pprev = tile2prv;
					tile2prv = tile2;
					path2 = path2.GetParent();
					if (path2) tile2 = path2.GetTile();
			}	
			// we've got a match on x or y, check if the direction is ok
			if ((tracks1 & AIRail.RAILTRACK_NE_SW) && 
				(tile2x == tile1x)){
				match = true;
			} else {

				if ((tracks1 & AIRail.RAILTRACK_NW_SE)&& 
					(tile2y == tile1y)){
					match = true;
				}	
			}
			match = true;
			if (!match) {
				// no match on x and y, restore second path's last match and
				//	forward the first path
				path2 = path2_sav;
				tile2pprev = tile2pp_sav;
				tile2prv = tile2prv_sav;
			} else {
				path2_sav = path2;
				tile2pp_sav = tile2pprev;
				tile2prv_sav = tile2prv;
				if ((AIMap.DistanceManhattan(tile1prv, tile2prv) < 12) ){
					local tiles_forward = AIMap.DistanceManhattan(tile1prv, tile2prv) + 2;
					local tmp_path = path2;
					local idx = 0;
				
					tiles_forward *= 2;
					while ((idx < (tiles_forward - 1)) && tmp_path){
						idx++;
						local tmp_tile;
						tmp_tile = tmp_path.GetTile();
						tracks1 = AIRail.GetRailTracks(tmp_tile);
						if ((tracks1 & AIRail.RAILTRACK_NE_SW) ||
							(tracks1 & AIRail.RAILTRACK_NW_SE)){
							idx += 1;
						}
						tmp_path = tmp_path.GetParent();
					}
					if(tile1pprev && tile1prv && tmp_path && tmp_path.GetTile() && tmp_path.GetParent()){
						pf.Plan([[tile1prv, tile1pprev]], [[tmp_path.GetTile(),tmp_path.GetParent().GetTile()]], [], 750);
						BuildPath(pf.path);
					}
					tmp_path = path1;
					idx = 0;
					while ((idx < (tiles_forward - 1)) && tmp_path){
						idx++;
						local tmp_tile;
						tmp_tile = tmp_path.GetTile();
						tracks1 = AIRail.GetRailTracks(tmp_tile);
						if ((tracks1 & AIRail.RAILTRACK_NE_SW) ||
							(tracks1 & AIRail.RAILTRACK_NW_SE)){
							idx += 1;
						}
						tmp_path = tmp_path.GetParent();
					}

					if(tile2pprev && tile2prv && tmp_path && tmp_path.GetTile() && tmp_path.GetParent()){
						pf.Plan([[tile2prv, tile2pprev]], [[tmp_path.GetTile(),tmp_path.GetParent().GetTile()]], [], 750);
						BuildPath(pf.path);
					}
					idx	= 0;
					tiles_forward = 15;

					while ((idx < tiles_forward ) && path1){
						idx++;
						tile1pprev = tile1prv;
						tile1prv = path1.GetTile();
						tracks1 = AIRail.GetRailTracks(tile1prv);
						if ((tracks1 & AIRail.RAILTRACK_NE_SW) ||
							(tracks1 & AIRail.RAILTRACK_NW_SE)){
							idx += 1;
						}
						path1 = path1.GetParent();
					}
				} // distance ok
			} // match	
		}

		if(path1){
			tile1pprev = tile1prv;
			tile1prv = path1.GetTile();
			path1  = path1.GetParent();
		}	
	}
}

function RailBuilder::SignalPath(path)
{
	local nof_safe_tiles = 0;
	local nof_tiles_since_sig = 0;
	local tile;
	local prv_tile;
	local pprv_tile;
	local last_signal = null;
	local last_signal_front = null;
	
	if (path) tile = path.GetTile();	
	
	while (path) {
		path = path.GetParent();
		if (path) {
			pprv_tile = prv_tile;
			prv_tile = tile;
			tile = path.GetTile();
			nof_safe_tiles++;
			nof_tiles_since_sig++;
			if (IsRailCrossing(tile)){
				if (nof_safe_tiles > 5) {
					if (AIRail.GetSignalType (prv_tile, tile) == AIRail.SIGNALTYPE_NONE){
						AIRail.BuildSignal(prv_tile, pprv_tile, AIRail.SIGNALTYPE_PBS);
						nof_tiles_since_sig = 0;				
					}	
				}	
				nof_safe_tiles = 0;		
			} 
			if (nof_safe_tiles > 5) {
				if (nof_tiles_since_sig > 5) {
					if (AIRail.GetSignalType (prv_tile, tile) == AIRail.SIGNALTYPE_NONE){
						AIRail.BuildSignal(prv_tile, pprv_tile, AIRail.SIGNALTYPE_PBS);
						nof_tiles_since_sig = 0;
						last_signal = prv_tile;
						last_signal_front = pprv_tile;
					}	
				}
			}
		}
	}
	if (nof_safe_tiles > nof_tiles_since_sig){
		if ((last_signal != null) && (last_signal_front!= null)){
			AIRail.RemoveSignal(last_signal, last_signal_front);
//			AISign.BuildSign(last_signal, "lsr" );
		}	
	}	
}

function RailBuilder::SignalPathRev(path)
{
	local nof_tiles_since_sig = 0;
	local tile = path.GetTile();
	local prv_tile = tile;
	local pprv_tile = tile;
	local last_signal = null;
	local last_signal_front = null;
	local pending_signal = false;
	local has_seen_crossing = false;
	
	while (path != null) {
		path = path.GetParent();
		if (path) {
			pprv_tile = prv_tile;
			prv_tile = tile;
			tile = path.GetTile();
			nof_tiles_since_sig++;
			
//			AISign.BuildSign(tile, " " + nof_tiles_since_sig + " " + IsRailCrossing(tile));
			
			if (IsRailCrossing(tile) 
				&& (!AIBridge.IsBridgeTile(tile)) 
				&& (!AITunnel.IsTunnelTile(tile))){
//				AISign.BuildSign(tile, " " + nof_tiles_since_sig + " " + IsRailCrossing(tile));
				if ((nof_tiles_since_sig <= 5) && (last_signal != null) ){
					if ((last_signal != null) && (last_signal_front!= null)){
						AIRail.RemoveSignal(last_signal, last_signal_front);
					}	
				}
				pending_signal = true;
				nof_tiles_since_sig = 0;				
				has_seen_crossing = true;
			} else if ((nof_tiles_since_sig > 5) || pending_signal) {
				local tmp_path = path.GetParent();
				local tmp_tile = 0;
				if (tmp_path) { 
					tmp_tile = tmp_path.GetTile();
				}	
				// only build signal if we've seen at least one crossing, to prevent deadlock
				// when trains enter the last section to the station, and the exiting train 
				// from the station has no alternative anymore
				if (has_seen_crossing){
					if (AIRail.GetSignalType (tile, tmp_tile) == AIRail.SIGNALTYPE_NONE){
						AIRail.BuildSignal(tile, tmp_tile, AIRail.SIGNALTYPE_PBS);
					}	
				}	
				pending_signal = false;
				last_signal = tile;
				last_signal_front = tmp_tile;
				nof_tiles_since_sig = 0;
			}
			
		}
	}	
}

