
/*------------------------------------------------------------------
|																	|
|	RailLine															|
|																	|
 ------------------------------------------------------------------*/

class RailLine
{
	towns              = null;
	stations           = null;
	depot              = null;
	vehicles           = null;
	date_last_vehicle  = 0;
	failed             = false;
	passenger_cargo_id = 0;
	mail_cargo_id      = 0;
	pending_vehicles   = 0;
	retry_build        = false;
	num_retries	       = 0;	
	line_len           = 0;
	state              = 0;
	pf 				   = 0;		
	first_path		   = null;
	second_path		   = null;	
	signals_present    = false;
	engine_build 	   = false;
	new_engine		   = 0;
	constructor(cargo_id, m_cargo_id, town_pair) {
		if (town_pair.len() == 2){
			towns     = town_pair;
		} else {
			AILog.Info("2 towns needed to create a Line");
			return null;
		}
		
		stations           = [null, null];
		depot              = [];
		vehicles           = [];
		failed             = false;
		date_last_vehicle  = 0;
		passenger_cargo_id = cargo_id;
		mail_cargo_id      = m_cargo_id;
		pending_vehicles   = 0;
		retry_build        = false;
		num_retries        = 0;
		line_len           = 0;
		state              = 0;
		pf 			       = RailPathFinder();
		first_path		   = null;
		second_path		   = null;
		signals_present    = false;
		engine_build 	   = false;
		new_engine		   = 0;
	}
	function AddVehicles();
	function BuildDepot();
	function CreateNewLine(town_pair);
	function ManageVehicles();
	function Controller();
}

function RailLine::Controller()
{
	AILog.Info("State " + state);
	switch (state){
		case 0:
			if (PlanStations()){
				state = 1;
			} else {
				state = 13;
			}	
			break;
		case 1:
			if (Cfo.HasMoney(50000)){
				local success = BuildStations();
				if(success == RailBuilder.SUCCESS) {
					state = 3;
					num_retries = 0;
				} else if (success == RailBuilder.REPLAN) {
					state = 0;
					num_retries++;
					if (num_retries > 3){
						state = 13;
						AILog.Info("Max retries: go to state " + state);
					}
				} else if (success == RailBuilder.RETRY_LATER) {
					state = 1;
				}	
				else state = 13;
			} else AIController.Sleep(10);
			
			break;
		
		case 3:
			if (Cfo.HasMoney(15000)) {
				if (PlanRoute(false)){
					state = 4;
				} else {
					if (pf.retry_find) { 
						if (num_retries < 3) {
							num_retries++;
						} else {
							state = 13;
						}
					} else {
						state = 13;
					}
				}	
			}  else AIController.Sleep(10);	
			break;
		case 4:
			if (Cfo.HasMoney(16000)) {
				if (BuildRoute()){
					state = 5;
					num_retries = 0;
				} else {
					if (pf.retry_find) { 
						if (num_retries < 3) {
							num_retries++;
							state = 3;
						} else {
							state = 13;
						}
					}			
				}		
			} else AIController.Sleep(10);	
			break;

		case 5:
			if (Cfo.HasMoney(15000)) {
				if (PlanRoute(true)){
					state = 6;
				} else {
					state = 13;
				}		
			} else AIController.Sleep(10);	
			break;
		case 6:
			if (Cfo.HasMoney(16000)) {
				if (BuildRoute()){
					AILog.Info("Buildroute " );
					local rb = RailBuilder();
					AILog.Info("Start switch point " );
					rb.MakeSwitchpoint(first_path, second_path);
					AILog.Info("goto state 7 " );
					num_retries = 0;
					state = 7;
				} else {
					AILog.Info("buildroute failed " );
					if (pf.retry_find) { 
						if (num_retries < 3) {
							num_retries++;
							state = 5;
							AILog.Info("buildroute failed back to state 5" );
						} else {
							state = 13;
						}
					}			
				}		
			} else AIController.Sleep(10);	
			break;			
		case 7:
			if (Cfo.HasMoney(3000)) {
				if (BuildDepot(false)) {
					AILog.Info("Depot built " + depot[0]);
					state = 8;
					num_retries = 0;
				}
				else {
					AILog.Info("Failed to build depot");
					if (num_retries < 3) {
						num_retries++;
					} else {
						state = 13;
					}
				}
			} else AIController.Sleep(10);	
			break;	
		case 8:
			if (Cfo.HasMoney(3000)) {
				if (BuildDepot(true)) {
					AILog.Info("Depot built " + depot[1]);
					state = 9;
					num_retries = 0;
				}
				else {
					AILog.Info("Failed to build depot");
					if (num_retries < 3) {
						num_retries++;
					} else {
						state = 13;
					}
				}
			} else AIController.Sleep(10);
			break;		
		case 9:
			if (Cfo.HasMoney(45000)){
				if (!signals_present){
					local rb = RailBuilder();
					rb.SignalPath(first_path);
					rb.SignalPathRev(second_path);
					signals_present = true;
				}	
				if (AddVehicles(depot[0], false)){
					state = 10;
				}
			} else AIController.Sleep(10);
			break;
		case 10:
			if (Cfo.HasMoney(45000)){
				if (AddVehicles(depot[1], true)){
					state = 11;
				}	
			} else AIController.Sleep(10);
			break;
		case 11:
			if (Cfo.HasMoney(44000)){
				if (ManageVehicles()){
					state = 13;
					num_retries = 0;
				}
			} else AIController.Sleep(10);
			break;
		default:
			break;
	}
	// Pay back unused money
	Cfo.ManageLoan();
	return (state < 11);
}

function RailLine::PlanStations()
{
	local town_idx     = 0;
	AILog.Info("Plan stations, line from: " + AITown.GetName(towns[0]) + " to " + AITown.GetName(towns[1]) + " " +  AICompany.GetBankBalance(AICompany.COMPANY_SELF));
	if(towns && (towns.len() == 2))
	{
		foreach (town in towns){
			if (!this.stations[town_idx]){
				local station_loc = TownManager.FindStationLocation(town, passenger_cargo_id, false);
				this.stations[town_idx] = RailStation(station_loc);
			}
			town_idx++;
		}
	} else {
		AILog.Info("Create new line failed: towns array not filled");
	}	
	if (this.stations[0]._location && this.stations[1]._location){
		AILog.Info("Plan stations: locations found");
		return true;
	} else {
		AILog.Info("Plan stations: locations not found");
		return false;
	}	
}

/*------------------------------------------------------------------*/
function RailLine::BuildStations()
{
	local town_idx  = 0;
	local rb      	= RailBuilder();
	
	foreach (town in towns){
		if (this.stations[town_idx] && this.stations[town_idx]._build_succes != RailBuilder.SUCCESS){
			AILog.Info("BuildStation: station nr " + town_idx );
			AILog.Info("BuildStation: " + AITown.GetName(towns[0]) + " to " + AITown.GetName(towns[1]) + " " +  AICompany.GetBankBalance(AICompany.COMPANY_SELF));
			this.stations[town_idx]._build_succes = rb.BuildRailStation(this.stations[town_idx]);
			if (this.stations[town_idx]._build_succes == RailBuilder.SUCCESS) {
				AILog.Info("Railstation " + town_idx + " built");
			} else {
				AILog.Info("New line: failed building station " + AIError.GetLastErrorString());
				return this.stations[town_idx]._build_succes;
			}
		}	
		town_idx++;
	}
	return RailBuilder.SUCCESS;
}


/*------------------------------------------------------------------*/
function RailLine::PlanRoute(second_line)
{
	local connected = true;
	local offset = 0;
	local target = [stations[1]._location - 1, stations[1]._location];

	local start; 
	
	if (second_line){
		start = stations[0].GetNextConnectionPoint();
	} else {
		start = stations[0].GetNewConnectionPoint();
	}	

	if (second_line){
		target = stations[1].GetNextConnectionPoint();
	} else {
		target = stations[1].GetNewConnectionPoint();
	}
	local swapped = false;
	if (stations[0]._connected_at_top == stations[1]._connected_at_top){
		swapped = true;
		if (second_line){
			target = stations[1].GetNewConnectionPoint();
		} else {
			target = stations[1].GetNextConnectionPoint();
		}	
	}
	local ignore;
	if (second_line){
		ignore = [];
	} else {
		if (swapped){
			ignore = [start[0] + AIMap.GetTileIndex(0, 1), target[0] - AIMap.GetTileIndex(0, 1)];	
		} else {
			ignore = [start[0] + AIMap.GetTileIndex(0, 1), target[0] + AIMap.GetTileIndex(0, 1)];	
		}
		ignore.append(ignore[0] + start[0] - start[1]);			
		ignore.append(ignore[1] + target[0] - target[1]);
	}	

	connected = pf.Plan([start], [target], ignore);
	if (!connected) {
		AILog.Info("Plan Route failed "+ AIError.GetLastErrorString());
	} else {
		if (second_line){
			second_path = pf.path;
		} else {
			first_path = pf.path;
		}	
	}
	

	return connected;
}

/*------------------------------------------------------------------*/
function RailLine::BuildRoute()
{
	local conn;
	local rb = RailBuilder();
	
	conn = (rb.BuildPath(pf.path) == RailBuilder.SUCCESS);

	if (conn){
		line_len = rb.path_len;

	} 
	else {
		AILog.Info("Build route failed"+ AIError.GetLastErrorString());
	}

	return conn;
}

/*------------------------------------------------------------------*/
function RailLine::AddVehicles(dep, reverse_order = false)
{
	local engine_started = false;
	if(!this.stations[0] || !this.stations[1]) {
		AILog.Info("AddVehicles nof stations incorrect");
		return false;
	}
	if (!AIRail.IsRailDepotTile(dep)){
		AILog.Info("AddVehicles: invalid depot parameter");
		return false;
	}
	local engine_model = null;
	local min_reliability = 85;

	local i;
	
	AILog.Info("Add vehicles to line from: " + AITown.GetName(towns[0]) + " to " + AITown.GetName(towns[1]) + " " +  AICompany.GetBankBalance(AICompany.COMPANY_SELF));

	local engine_list = AIEngineList(AIVehicle.VT_RAIL);
	if (!engine_build){
		engine_list.Valuate(AIEngine.GetRailType);
		engine_list.KeepValue(AIRail.GetCurrentRailType());  // todo: store rt with line

		foreach (eng,dum in engine_list){
//			AILog.Info("engine : " + AIEngine.GetName(eng));
		}
		local balance = AICompany.GetBankBalance(AICompany.COMPANY_SELF);
		engine_list.Valuate(AIEngine.GetPrice);
		engine_list.KeepBelowValue(balance);
	
		pending_vehicles = 1; 

		if (!engine_list.Count()) {
			AILog.Info("AddVehicles: failed to find an engine to build");
			return false;
		}

		engine_model = engine_list.Begin();

		// Buy engine

		local town_idx = 0;	
		new_engine = AIVehicle.BuildVehicle(dep, engine_model);

		if (AIVehicle.IsValidVehicle(new_engine)) { 
			this.vehicles.append(new_engine);

			AIOrder.AppendOrder(new_engine, dep, AIOrder.AIOF_SERVICE_IF_NEEDED);
			if (reverse_order) town_idx = (town_idx + 1) % 2;
			AIOrder.AppendOrder(new_engine, this.stations[town_idx]._location, AIOrder.AIOF_NONE); 
			town_idx = (town_idx + 1) % 2;
			AIOrder.AppendOrder(new_engine, this.stations[town_idx]._location, AIOrder.AIOF_NONE);
		}
		else {
			AILog.Info("Buy engine failed "+ AIError.GetLastErrorString() + " " + dep);
			AILog.Info("failed engine : " + AIEngine.GetName(engine_model));
			return false;
		}
	}	
	engine_build = true;		

	/* we've got an engine, now add some wagons to the train */	
	engine_list = AIEngineList(AIVehicle.VT_RAIL);
	engine_list.Valuate(AIEngine.CanRunOnRail, AIRail.GetCurrentRailType());
	engine_list.KeepValue(1); 
	engine_list.Valuate(AIEngine.IsWagon);
	engine_list.KeepValue(1);
	
//	AILog.Info("wagon list count, no cargo check done: " + engine_list.Count());
//	foreach (eng,dum in engine_list){
//		AILog.Info("wagon : " + AIEngine.GetName(eng));
//	}
	engine_list.Valuate(AIEngine.GetCargoType)
	engine_list.KeepValue(passenger_cargo_id); 

	AILog.Info("wagon list count: " + engine_list.Count());
//	foreach (eng,dum in engine_list){
//		AILog.Info("engine : " + AIEngine.GetName(eng));
//	}
	local balance = AICompany.GetBankBalance(AICompany.COMPANY_SELF);
	engine_list.Valuate(AIEngine.GetPrice);
	engine_list.KeepBelowValue(balance);

	engine_list.Valuate(AIEngine.GetCapacity)
	engine_list.KeepTop(1);	
	if (!engine_list.Count()) {
		AILog.Info("no wagons to add" + engine_list.Count());
		return false;
	}	
	local wagon_model = engine_list.Begin();		
	
	do{
		local new_wagon = AIVehicle.BuildVehicle(dep, wagon_model);
		AIVehicle.MoveWagon(new_wagon, 0, new_engine, 0);
	} while (AIVehicle.GetLength(new_engine) < (16 * 4 + 8));   /* todo, make stationlen a variable */


	// new we've got pax wagons, let's add some mail wagons	
	
	local mail_engine_list = AIEngineList(AIVehicle.VT_RAIL);

	mail_engine_list.Valuate(AIEngine.CanRunOnRail, AIRail.GetCurrentRailType());
	mail_engine_list.KeepValue(1); 
	mail_engine_list.Valuate(AIEngine.IsWagon);
	mail_engine_list.KeepValue(1);
	
//	AILog.Info("wagon list count, no cargo check done: " + engine_list.Count());
//	foreach (eng,dum in mail_engine_list){
//		AILog.Info("mail wagon : " + AIEngine.GetName(eng));
//	}
	mail_engine_list.Valuate(AIEngine.GetCargoType)
	mail_engine_list.KeepValue(mail_cargo_id); 

	AILog.Info("mail_wagon list count: " + mail_engine_list.Count());
//	foreach (eng,dum in engine_list){
//		AILog.Info("engine : " + AIEngine.GetName(eng));
//	}
	balance = AICompany.GetBankBalance(AICompany.COMPANY_SELF);
	mail_engine_list.Valuate(AIEngine.GetPrice);
	mail_engine_list.KeepBelowValue(balance);

	mail_engine_list.Valuate(AIEngine.GetCapacity)
	mail_engine_list.KeepTop(1);	
	if (!mail_engine_list.Count()) {
		// not fatal, we still got the pax wagons
		AILog.Info("no mail_wagons to add" + mail_engine_list.Count());
	} else {	
		local wagon_model = mail_engine_list.Begin();		
		do{
			local new_wagon = AIVehicle.BuildVehicle(dep, wagon_model);
			AIVehicle.MoveWagon(new_wagon, 0, new_engine, 0);
		} while (AIVehicle.GetLength(new_engine) < (16 * 5));   /* todo, make stationlen a variable */
	}

	AIVehicle.StartStopVehicle(new_engine);
	engine_started  = true;	
	engine_build = false;  // init for next veh	
	AILog.Info("Train vehicles pending:" + pending_vehicles);
	return (engine_started);

}



/*------------------------------------------------------------------*/
function RailLine::BuildDepot(second_depot)
{
	local aitile	= 0;
	local found		= false;
	local from   	= null;
	local fromfrom  = null;
	local to		= null;
	local toto		= null;
	local n			= 0;
	local path		= null;
	
	if (second_depot){
		path = second_path;
	} else {
		path = first_path;
	}	
	while (path != null){
		fromfrom = from;
		from = aitile;
		aitile = to; 
		to = toto;
		toto = path.GetTile();
		n++;
		// filter out the tiles too close to the station,
		// and the short diagonal tracks (give problems on forbid 90 degrees setting)
		// TODO: place the depot a little further away from the track for short tracks
		if ((n > 20) 
			&& ((aitile - from) == (to - aitile))
			&& ((from - fromfrom) == (aitile - from))
			&& ((to - aitile) == (toto - to))) {
			local adjacentTiles = Tile.GetAdjacentTiles(aitile);

			/*Loop through all adjacent tiles*/
			for(local tile2 = adjacentTiles.Begin(); 
				adjacentTiles.HasNext() && !found; 
				tile2 = adjacentTiles.Next()) {
				
				if ( !AIRail.IsRailTile(tile2) 
				  && !AITile.GetSlope(tile2)  
				  && !AIRail.IsRailStationTile(tile2) 
				  &&  AITile.IsBuildable(tile2)) {
					local rb  = RailBuilder();
					local success = false;
					success = rb.BuildDepot(from, aitile, to, tile2);
					found = (success == RailBuilder.SUCCESS); 
					if (found) depot.append(tile2);
				}
			}
			if (found) {
				break;
			}
		}
		path = path.GetParent();
	}
	
	if (found) {
		return depot[depot.len() - 1];	
	}
	else {
		return null;
	}
}



/*------------------------------------------------------------------*/
function RailLine::ManageVehicles()
{
	local max_veh_on_line;
	local max_veh_reached = false;
	
	local curr_station = AIStation.GetStationID(this.stations[0]._location);
	local veh_list_station = AIVehicleList_Station(curr_station);
	local veh_on_line = veh_list_station.Count();		

	max_veh_on_line  = 2 + this.line_len / 40;
	if (veh_on_line < max_veh_on_line) {
		local second_station = AIStationList_Vehicle(veh_list_station.Begin()).Begin();
		local waiting = AIStation.GetCargoWaiting(curr_station,
											  	  passenger_cargo_id) +
													AIStation.GetCargoWaiting(AIStation.GetStationID(second_station),
													passenger_cargo_id);
		local rating0 = AIStation.GetCargoRating(curr_station, passenger_cargo_id);
		local rating1 = AIStation.GetCargoRating(second_station, passenger_cargo_id);

		if ((waiting > 325) || (rating0 < 75) || (rating1 < 75)) {
			if (AIDate.GetCurrentDate() - this.date_last_vehicle > 50) {
				if (veh_list_station.Begin() && this.depot[0]) {
					this.pending_vehicles = this.pending_vehicles + waiting / 200;
				}
			}
		}
	}
//	AILog.Info("Manage vehicles: pending , max " + this.pending_vehicles + ", " + max_veh_on_line + ", " + this.vehicles.len());
	if (this.vehicles.len() + this.pending_vehicles >= max_veh_on_line){
		this.pending_vehicles = max_veh_on_line - this.vehicles.len();
	}	
			
	if ((AIDate.GetCurrentDate() - this.date_last_vehicle) > 10){
		while (this.pending_vehicles) {
			if (AddVehicles(this.depot[0], false)){
				this.date_last_vehicle = AIDate.GetCurrentDate();
				if (this.vehicles.len() >= max_veh_on_line){
					max_veh_reached = true;
					AILog.Info("Manage vehicles: max veh reached " + max_veh_on_line);
				}	
				this.pending_vehicles = this.pending_vehicles - 1;
			} else {
				return false;  // probably out of money, retry later
			}	
		}
	}

	return (max_veh_reached);
}

