
class RoadBuilder
{
	SUCCESS        = 0;
	RETRY_LATER    = 1;
	REPLAN         = 2;
	FAIL           = 3;
	path_len       = 0;
	constructor() {
		SUCCESS        = 0;
		RETRY_LATER    = 1;
		REPLAN         = 2;
		FAIL           = 3;
		path_len       = 0;
	}
	function BuildPath( path );
}


function RoadBuilder::BuildPath( path )
{
	local build_success = false;
	local retry_find    = false;
	path_len     	    = 0;	
	
	while (path != null) {
		path_len = path_len + 1;	
		if (path.GetParent() != null) {
			local parnt = path.GetParent().GetTile();
			if (!AIRoad.AreRoadTilesConnected(path.GetTile(), parnt) 
				&& (!AIBridge.IsBridgeTile(path.GetTile()) 
					|| AIBridge.GetOtherBridgeEnd(path.GetTile()) != parnt)) {
				local retry_build = false;
				local build_retries = 0;	 
				do {
					retry_build = false;
					if (AIMap.DistanceManhattan(path.GetTile(), parnt) == 1 ) {
						build_success = AIRoad.BuildRoad(path.GetTile(), parnt);
					}
					else {
						AILog.Info("PathFinder: build bridge");
						/* Build a bridge or tunnel. */
						if (!AIBridge.IsBridgeTile(path.GetTile()) && !AITunnel.IsTunnelTile(path.GetTile())) {
							/* If it was a road tile, demolish it first. Do this to work around expended roadbits. */
							if (AIRoad.IsRoadTile(path.GetTile())) AITile.DemolishTile(path.GetTile());
							if (AITunnel.GetOtherTunnelEnd(path.GetTile()) == parnt) {
								build_success = AITunnel.BuildTunnel(AIVehicle.VT_ROAD, path.GetTile());
								if (!build_success) {
									AILog.Info("Build tunnel failed" + AIError.GetLastErrorString());
								}
							} else {
								local bridge_list = AIBridgeList_Length(AIMap.DistanceManhattan(path.GetTile(), parnt) + 1);
								bridge_list.Valuate(AIBridge.GetMaxSpeed);
								
								bridge_list.Sort(AIAbstractList.SORT_BY_VALUE, false);
								build_success = AIBridge.BuildBridge(AIVehicle.VT_ROAD, bridge_list.Begin(), path.GetTile(), parnt)
								if (!build_success) {
									AILog.Info("Build bridge failed" + AIError.GetLastErrorString());
								}
							}
						}
					}
					if (!build_success) {
						local err = AIError.GetLastError();
						switch (err) {
							/* ignore these errors */
							case AIError.ERR_NONE: 	
							case AIError.ERR_ALREADY_BUILT: 
								AILog.Info("Build road:"+ AIError.GetLastErrorString());
								build_success = true;
							break;
							case AIError.ERR_NOT_ENOUGH_CASH:  
								AILog.Info("Build road failed (no cash)"+ AIError.GetLastErrorString());
								return RETRY_LATER;
							break;
							case AIError.ERR_VEHICLE_IN_THE_WAY: 
								AILog.Info("Build road failed (retry)"+ AIError.GetLastErrorString());	
								AIController.Sleep(1);
							break;
							case AIError.ERR_AREA_NOT_CLEAR: 	
								AILog.Info("Build road failed (demolish + retry)"+ AIError.GetLastErrorString());	
								retry_find = !AITile.DemolishTile(path.GetTile());
								break;
							default:
								AILog.Info("Build road failed (replan)"+ AIError.GetLastErrorString());	
								retry_find = true;
								break;
						}
					} 	
					build_retries++;					
				} while (!build_success && build_retries < 80 && !retry_find);
			}
			else build_success = true;
  		}
		path = path.GetParent();
	}
	
	if (build_success)
		return SUCCESS;
	else if (retry_find) 
		return REPLAN;
	else
		return FAIL;	
}



function RoadBuilder::BuildRoadStation( tile, tile2, passengerstation )
{

	local success = AIRoad.BuildRoad(tile, tile2);

	if (!success){
		local err = AIError.GetLastError();
		switch (err) {
			/* ignore these errors */
			case AIError.ERR_NONE: 	
			case AIError.ERR_ALREADY_BUILT: 
				AILog.Info("Build road:" + AIError.GetLastErrorString());
			break;
			case AIError.ERR_NOT_ENOUGH_CASH:  
				AILog.Info("Build road station road failed (no cash)" + AIError.GetLastErrorString());
				return RETRY_LATER;
			break;
			default:
				AISign.BuildSign(tile, "br fail");
				AISign.BuildSign(tile2, "br fail");
				AILog.Info("Build road station road failed " + AIError.GetLastErrorString());
				return FAIL;
			break;
		}
	}
	
	local truck;
	if (passengerstation){
		truck = AIRoad.ROADVEHTYPE_BUS;  
	} else {
		truck = AIRoad.ROADVEHTYPE_TRUCK;  
	}	
	
	local adjacent = AIStation.STATION_NEW;
	success = AIRoad.BuildRoadStation(tile, tile2, truck, adjacent);
	if (!success){
		AILog.Info("Build busstop failed " + AIError.GetLastErrorString());
		local err = AIError.GetLastError();
		switch (err) {
			/* ignore these errors */
			case AIError.ERR_NONE: 	
			case AIError.ERR_ALREADY_BUILT: 
				AILog.Info("Build road:" + AIError.GetLastErrorString());
				return SUCCESS;
			break;
			case AIError.ERR_NOT_ENOUGH_CASH:  
				AILog.Info("Build road station failed (no cash)" + AIError.GetLastErrorString());
				return RETRY_LATER;
			break;
			default:
				AILog.Info("Build road station failed " + AIError.GetLastErrorString());
				return FAIL;
			break;
		}
	} else {
		return SUCCESS;
	}	
}

function RoadBuilder::BuildDepot( tile, tile2 )
{
	local rb  = RoadBuilder();
	local success = AIRoad.BuildRoad(tile, tile2);									
	
	if (!success){
		local err = AIError.GetLastError();
		switch (err) {
			/* ignore these errors */
			case AIError.ERR_NONE: 	
			case AIError.ERR_ALREADY_BUILT: 
				AILog.Info("Build Depot road:" + AIError.GetLastErrorString());
			break;
			case AIError.ERR_NOT_ENOUGH_CASH:  
				AILog.Info("Build depot road failed (no cash)" + AIError.GetLastErrorString());
				return RETRY_LATER;
			break;
			default:
				AILog.Info("Build Depot road failed " + AIError.GetLastErrorString());
				return FAIL;
			break;
		}
	}	

	success = AIRoad.BuildRoadDepot(tile, tile2)
	
	if (!success){
		local err = AIError.GetLastError();
		switch (err) {
			/* ignore these errors */
			case AIError.ERR_NONE: 	
			case AIError.ERR_ALREADY_BUILT: 
				AILog.Info("Build Depot:" + AIError.GetLastErrorString());
				return SUCCESS;
			break;
			case AIError.ERR_NOT_ENOUGH_CASH:  
				AILog.Info("Build Depot failed (no cash)" + AIError.GetLastErrorString());
				return RETRY_LATER;
			break;
			default:
				AILog.Info("Build road station failed " + AIError.GetLastErrorString());
				return FAIL;
			break;
		}
	} else {
		return SUCCESS;
	}	
}
