

/*------------------------------------------------------------------
|																	|
|	TownManager														|
|																	|
 ------------------------------------------------------------------*/
class TownManager
{
	constructor() {

	}

	function FindLineBusStopLocation(town, pass_cargo_id, estimate);
	function FindStationLocation(town, pass_cargo_id, estimate);
	function EstimateAcceptance(town);
}


/*------------------------------------------------------------------*/
function TownManager::FindLineBusStopLocation(town, pass_cargo_id, estimate)
{
	local tl = AITileList();
	local aitile = null;
	local tile = AITown.GetLocation(town);
	local found = false;
	local ret_tile = null;
	local sl = AIStationList(AIStation.STATION_BUS_STOP);
	local rsl = AIStationList(AIStation.STATION_TRAIN);
	
	tl.AddRectangle(tile + AIMap.GetTileIndex(-8, -8), tile + AIMap.GetTileIndex(8, 8));

	sl.Valuate(AIStation.GetDistanceManhattanToTile, tile);
	sl.KeepBetweenValue(0, 16);
	
	/* remove tiles around our own bus stations */
	for (local st_it = sl.Begin(); sl.HasNext(); st_it = sl.Next()){
		tl.RemoveRectangle(AIStation.GetLocation(st_it) + AIMap.GetTileIndex(-4, -4), AIStation.GetLocation(st_it) + AIMap.GetTileIndex(4, 4));
	}
	/* remove tiles around our own train stations */
	for (local st_it = rsl.Begin(); rsl.HasNext(); st_it = rsl.Next()){
		tl.RemoveRectangle(AIStation.GetLocation(st_it) + AIMap.GetTileIndex(-8, -8), AIStation.GetLocation(st_it) + AIMap.GetTileIndex(8, 8));
	}
	/* remove all tiles that are already covered by opponents stations */
	local tl2 = AITileList();
	
	tl2.AddList(tl);
	tl2.Valuate(AIRoad.IsRoadStationTile);
	tl2.KeepValue(1);

	for (local rstl = tl2.Begin(); tl2.HasNext() ; rstl = tl2.Next()){
		/* Keep our own stations a bit apart */
		if (AITile.GetOwner(rstl) != AICompany.ResolveCompanyID(AICompany.COMPANY_SELF)) {
			tl.RemoveRectangle(rstl + AIMap.GetTileIndex(-1, -1), rstl + AIMap.GetTileIndex(1, 1));
		}
	}	 

	if (tl.Count()) {
		/* find all tiles that are next to a road tile */
		tl.Valuate(AIRoad.GetNeighbourRoadCount);	
		tl.KeepAboveValue(0);

		if (tl.Count()) {
			/* find all tiles that are not road */
			tl.Valuate(AIRoad.IsRoadTile);	
			tl.KeepValue(0);
			if (tl.Count()) {
				/* find all tiles that are not sloped */
				tl.Valuate(AITile.GetSlope);	
				tl.KeepValue(0);
				if (tl.Count()) {
					tl.Valuate(AITile.GetCargoAcceptance, pass_cargo_id,
								 1, 1, 
								 AIStation.GetCoverageRadius (AIStation.STATION_BUS_STOP));
					for (aitile = tl.Begin(); tl.HasNext(); aitile = tl.Next()){
//						AISign.BuildSign(aitile,"ex"+tl.GetValue(aitile));
						if (tl.GetValue(aitile) >= 15){
  							found = AITile.IsBuildable(aitile);
							if (estimate) {
								found = true;
							}	
							if (!found)	{
								found = AITile.DemolishTile(aitile);
							}
							if (found) {
								ret_tile = aitile;
								break;
							}	
						}	
						else {
//							AILog.Info("Find busstop location, acceptance too low " + tl.GetValue(aitile));
						}
					}
				}
				else {
					AILog.Info("Find busstop location, no unsloped tiles present");
				}
			}
			else {
				AILog.Info("Find busstop location, no tiles that are not road present");
			}
		}
		else {
			AILog.Info("Find busstop location, no tiles next to road present");
		}
	}
	

	
	if (found) {
//		AISign.BuildSign(aitile,"ex"+tl.GetValue(aitile));
		return ret_tile;	
	}	
	else {
		return null;	
	}	
}

/*------------------------------------------------------------------*/
function TownManager::FindMailTruckStopLocation(tile, mail_cargo_id, estimate)
{
	local tl = AITileList();
	local aitile = null;
	local found = false;
	local ret_tile = null;
	
	tl.AddRectangle(tile + AIMap.GetTileIndex(-3, -3), tile + AIMap.GetTileIndex(3, 3));
	
	/* remove all tiles that are already covered by a station */
	local tl2 = AITileList();

	tl2.AddList(tl);
	tl2.Valuate(AIRoad.IsRoadStationTile);
	tl2.KeepValue(1);

	for (local rstl = tl2.Begin(); tl2.HasNext() ; rstl = tl2.Next()){
		tl.RemoveRectangle(rstl + AIMap.GetTileIndex(-1, -1), rstl + AIMap.GetTileIndex(1, 1));
	}	 

	if (tl.Count()) {
		/* find all tiles that are next to a road tile */
		tl.Valuate(AIRoad.GetNeighbourRoadCount);	
		tl.KeepAboveValue(0);

		if (tl.Count()) {
			/* find all tiles that are not road */
			tl.Valuate(AIRoad.IsRoadTile);	
			tl.KeepValue(0);
			if (tl.Count()) {
				/* find all tiles that are not sloped */
				tl.Valuate(AITile.GetSlope);	
				tl.KeepValue(0);
				if (tl.Count()) {
					tl.Valuate(AITile.GetCargoAcceptance, mail_cargo_id,
								 1, 1, 
								 AIStation.GetCoverageRadius (AIStation.STATION_TRUCK_STOP));
					for (aitile = tl.Begin(); tl.HasNext(); aitile = tl.Next()){
//						AISign.BuildSign(aitile,"ex"+tl.GetValue(aitile));
						if (tl.GetValue(aitile) >= 6){
  							found = AITile.IsBuildable(aitile);
							if (!found)	{
								found = AITile.DemolishTile(aitile);
							}
							if (found) {
								ret_tile = aitile;
								break;
							}	
						}	
						else {
							AILog.Info("Find mailtruckstop location, acceptance too low " + tl.GetValue(aitile));
						}
					}
				}
				else {
					AILog.Info("Find mailtruckstop location, no unsloped tiles present");
				}
			}
			else {
				AILog.Info("Find mailtruckstop location, no tiles that are not road present");
			}
		}
		else {
			AILog.Info("Find mailtruckstop location, no tiles next to road present");
		}
	}
	

	
	if (found) {
//		AISign.BuildSign(aitile,"ex"+tl.GetValue(aitile));
		return ret_tile;	
	}	
	else {
		return null;	
	}	
}

/*------------------------------------------------------------------*/
function TownManager::EstimateAcceptanceBus(town, passenger_cargo_id)
{
	local new_location = TownManager.FindLineBusStopLocation(town,
																passenger_cargo_id, true);
		
	local acceptance = 0;
	if (new_location){

		acceptance = AITile.GetCargoAcceptance(new_location, passenger_cargo_id, 1, 1, 
													 AIStation.GetCoverageRadius (AIStation.STATION_BUS_STOP));					

	}												 
	return [acceptance, new_location];											 

}

/*------------------------------------------------------------------*/
function TownManager::EstimateAcceptanceTrain(town, passenger_cargo_id)
{
	local new_location = TownManager.FindStationLocation(town,
														passenger_cargo_id, true);
		
	local acceptance = 0;
	if (new_location){

		acceptance = AITile.GetCargoAcceptance(new_location, passenger_cargo_id, 5, 2, 
													 AIStation.GetCoverageRadius (AIStation.STATION_TRAIN));					

	}												 
	return [acceptance, new_location];											 

}

/*------------------------------------------------------------------*/
function TownManager::FindStationLocation(town, pass_cargo_id, estimate)
{
	local tl = AITileList();
	local aitile = null;
	local tile = AITown.GetLocation(town);
	local found = false;
	local ret_tile = null;

	local sl = AIStationList(AIStation.STATION_TRAIN);

	sl.Valuate(AIStation.GetDistanceManhattanToTile, tile);
	sl.KeepBetweenValue(0, 16);
	
	tl.AddRectangle(tile + AIMap.GetTileIndex(-8, -8), tile + AIMap.GetTileIndex(8, 8));
	/* remove tiles around our own stations */
	for (local st_it = sl.Begin(); sl.HasNext(); st_it = sl.Next()){
		tl.RemoveRectangle(AIStation.GetLocation(st_it) + AIMap.GetTileIndex(-12, -12), 
						   AIStation.GetLocation(st_it) + AIMap.GetTileIndex(12, 12));
	}
	if (tl.Count()) {
		tl.Valuate(AITile.GetCargoAcceptance, pass_cargo_id,
					5, 2, 
					AIStation.GetCoverageRadius (AIStation.STATION_TRAIN));
		for (aitile = tl.Begin(); tl.HasNext(); aitile = tl.Next()){
			if (tl.GetValue(aitile) >= 15){
				found = AITile.IsBuildableRectangle(aitile, 7, 2);
						
				if (estimate) {
//					found = true;
				}	
				if (found) {
					local tl2 = AITileList();
					tl2.AddRectangle(aitile, aitile + AIMap.GetTileIndex(4, 1));
					tl2.Valuate(AITile.GetSlope);	
					tl2.KeepValue(0);
					tl2.Valuate(AIRail.IsRailTile);	
					tl2.KeepValue(0);
					tl2.Valuate(AIRoad.IsRoadTile);	
					tl2.KeepValue(0);
					if (tl2.Count() == 10){
						if (!estimate){
							for (local tl2tile = tl2.Begin(); tl2.HasNext(); tl2tile = tl2.Next()){
								AITile.DemolishTile(tl2tile);
							}
//							AISign.BuildSign(aitile,"ts"+tl.GetValue(aitile));
						}				
						ret_tile = aitile;
						break;
					}
				}	
			}	
			else {
//				AILog.Info("Find station location, acceptance too low " + tl.GetValue(aitile));
			}
		}
	} else {
		AILog.Info("Find station location, no tiles ");
	}	
	if (found) {
		return ret_tile;	
	}	
	else {
		return null;	
	}	
}
