/*------------------------------------------------------------------
|																	|
|	Convoy															|
|																	|
 -------------------------------------------------------------------- 
|
| Author : GeekToo
| Version: 12
|
|
 --------------------------------------------------------------------*/
import("pathfinder.road", "RoadPathFinder", 3);
//import("pathfinder.rail", "RPathFinder", 2);

require("Cfo.nut");
require("Line.nut");
require("PathFinder.nut");
require("pf.rail/pfr.nut");
require("RailPathFinder.nut");
require("RailBuilder.nut");
require("RailLine.nut");
require("RailStation.nut");
require("RoadBuilder.nut");
require("RoutePlanner.nut");
require("Tile.nut");
require("TownManager.nut");


/*----------------------------------------------------------------------
|																		|
|	TConvoyAI															|
|																		|
 ----------------------------------------------------------------------*/
class TConvoyAI extends AIController {
	agressive			 = true;
	company            = null;
	stop               = false;
	vehicle_to_depot   = {};
	route_planner      = null;
	lines              = [];
	rlines             = [];
	name               = null;
	passenger_cargo_id = 0;
	mail_cargo_id      = 0;

	constructor() {
		this.company   = AICompany();
		route_planner  = null;
		name = null;

	}

	function Start();

	function SetCompanyName();
	function Load(version, data);
	function Save();
}



/*------------------------------------------------------------------*/
function TConvoyAI::Start()
{
	this.SetCompanyName();
	this.name = AICompany.GetName(AICompany.COMPANY_SELF);
	agressive = GetSetting("Agressive");
	AILog.Info(this.name + " (12) is starting");
	AILog.Info("Setting of agressive is: " + agressive);
	AIRoad.SetCurrentRoadType(AIRoad.ROADTYPE_ROAD);
	
	local rail_type_list = AIRailTypeList();
	AIRail.SetCurrentRailType(rail_type_list.Begin());
	
	local cargo_list = AICargoList();

	cargo_list.Valuate(AICargo.HasCargoClass, AICargo.CC_PASSENGERS);
	cargo_list.KeepValue(1);
	if (cargo_list.Count() == 0) {
	/* There is no passenger cargo, so adding buses is useless. */
		this.passenger_cargo_id = null;
		return;
	}
	if (cargo_list.Count() > 1) {
		local town_list = AITownList();
		town_list.Valuate(AITown.GetPopulation);
		town_list.Sort(AIAbstractList.SORT_BY_VALUE, false);
		local best_cargo = null;
		local best_cargo_acceptance = 0;
		foreach (cargo, dummy in cargo_list) {
			local acceptance = AITile.GetCargoAcceptance(AITown.GetLocation(town_list.Begin()), cargo, 1, 1, 5);
			if (acceptance > best_cargo_acceptance) {
				best_cargo_acceptance = acceptance;
				best_cargo = cargo;
			}
		}
			
		this.passenger_cargo_id = best_cargo;
	} else {
		this.passenger_cargo_id = cargo_list.Begin();
	}

	AICompany.SetAutoRenewMonths(-2);
	AICompany.SetAutoRenewStatus(true);	
	
	cargo_list.Clear();
	cargo_list = AICargoList();
	
	cargo_list.Valuate(AICargo.HasCargoClass, AICargo.CC_MAIL);
	cargo_list.KeepValue(1);
	if (cargo_list.Count() == 0) {
	/* There is no mail cargo, so adding mail trucks is useless. */
		this.mail_cargo_id = null;
	}
	if (cargo_list.Count() > 1) {
		local town_list = AITownList();
		town_list.Valuate(AITown.GetPopulation);
		
		town_list.Sort(AIAbstractList.SORT_BY_VALUE, false);
		local best_cargo = null;
		local best_cargo_acceptance = 0;
		foreach (cargo, dummy in cargo_list) {
			local acceptance = AITile.GetCargoAcceptance(AITown.GetLocation(town_list.Begin()), cargo, 1, 1, 5);
			if (acceptance > best_cargo_acceptance) {
				best_cargo_acceptance = acceptance;
				best_cargo = cargo;
			}
		}
		this.mail_cargo_id = best_cargo;
	} else {
		this.mail_cargo_id = cargo_list.Begin();
	}
	
	local engine_list = AIEngineList(AIVehicle.VT_ROAD);
	
	engine_list.Valuate(AIEngine.GetRoadType);
	engine_list.KeepValue(AIRoad.ROADTYPE_ROAD);

	engine_list.Valuate(AIEngine.GetCargoType);
	engine_list.KeepValue(this.passenger_cargo_id);
	
	if (engine_list.Count() == 0) {
		AILog.Info("Stopping because no road passenger vehicles are available");
		return;
	}
	route_planner = RoutePlanner(this.passenger_cargo_id);
	route_planner.InitTownList();

	while(!this.stop)
	{
		this.Sleep(1);

		local nof_lines_constructing = 0;
		local nof_rlines_constructing = 0;
//		foreach ( line_it in lines) {
//			if (line_it.Controller()){
//				nof_lines_constructing = nof_lines_constructing + 1;
//			}	
//		}
		foreach ( line_it in rlines) {
			if (line_it.Controller()){
				nof_rlines_constructing = nof_rlines_constructing + 1;
			}	
		}
		if ((nof_lines_constructing < 1) && (nof_rlines_constructing < 1)){
			local towns;
			local rtowns; 
			towns = this.route_planner.FindUnusedTowns(agressive, true);
			rtowns = this.route_planner.FindUnusedTowns(agressive, false);
			if(towns && (towns.len() == 2)) {
//				local line = Line(this.passenger_cargo_id, this.mail_cargo_id, towns);
//				lines.append(line);
//				AILog.Info("appended line");
			}	
			if(rtowns && (rtowns.len() == 2)) {
				local rail_line = RailLine(this.passenger_cargo_id, this.mail_cargo_id, rtowns);
				rlines.append(rail_line);
				AILog.Info("appended rline");
			}
		}	
	}	
}


/*------------------------------------------------------------------*/
function TConvoyAI::SetCompanyName()
{
	if (!AICompany.SetName("Convoy")) {
		local i = 2;
		while (!AICompany.SetName("Convoy#" + i)) {
			i++;
		}
	}
}


/*------------------------------------------------------------------*/
/*	Just a dummy, to prevent warnings						 		*/ 	
/*------------------------------------------------------------------*/
function TConvoyAI::Save()
{
	local table = {dummy = false};
	return table;
}

/*------------------------------------------------------------------*/
function TConvoyAI::Load(version, data)
{
  
}
